from .base58 import *
from .bip32 import *
from .chain import *
from .coin import *
from .hashes import *
from .keys import *
from .misc import *
from .packing import *
from .script import *
from .tx import *
from .work import *

_version_str = '0.1.1'
_version = tuple(int(part) for part in _version_str.split('.'))

__all__ = sum((
    base58.__all__,
    bip32.__all__,
    chain.__all__,
    coin.__all__,
    hashes.__all__,
    keys.__all__,
    misc.__all__,
    packing.__all__,
    script.__all__,
    tx.__all__,
    work.__all__,
), ())



# Copyright (c) 2018, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

__all__ = (
    'Bitcoin', 'BitcoinTestnet', 'BitcoinScalingTestnet', 'Coin', 'Header', 'all_coins',
)

import attr

from bitcoinx.hashes import double_sha256, hash_to_hex_str, hash_to_value
from bitcoinx.packing import unpack_header, unpack_le_uint32
from bitcoinx.work import (
    bits_to_work, bits_to_target, required_bits_mainnet, required_bits_testnet,
    required_bits_scaling_testnet
)


@attr.s(slots=True)
class Header(object):
    version = attr.ib()
    prev_hash = attr.ib()
    merkle_root = attr.ib()
    timestamp = attr.ib()
    bits = attr.ib()
    nonce = attr.ib()

    # Extra metadata
    hash = attr.ib()
    raw = attr.ib()
    height = attr.ib()

    def work(self):
        return bits_to_work(self.bits)

    def target(self):
        return bits_to_target(self.bits)

    def hash_value(self):
        return hash_to_value(self.hash)

    def hex_str(self):
        return hash_to_hex_str(self.hash)

    def difficulty(self):
        return Bitcoin.bits_to_difficulty(self.bits)

    def __str__(self):
        return (f'Header(version=0x{self.version:x}, prev_hash={hash_to_hex_str(self.prev_hash)}, '
                f'merkle_root={hash_to_hex_str(self.merkle_root)}, timestamp={self.timestamp}, '
                f'bits=0x{self.bits}, nonce={self.nonce}, hash={self.hex_str()}, '
                f'height={self.height})')


class Coin(object):

    def __init__(self, name, genesis_header, required_bits, P2PKH_verbyte, WIF_byte,
                 xpub_verbytes, xprv_verbytes):
        self.name = name
        self.genesis_header = bytes.fromhex(genesis_header)
        self.genesis_bits = self.header_bits(self.genesis_header)
        self.max_target = bits_to_target(self.genesis_bits)
        # Signature:  def required_bits(self, headers, chain, height, timestamp=None)
        self.required_bits = required_bits
        self.P2PKH_verbyte = P2PKH_verbyte
        self.WIF_byte = WIF_byte
        self.xpub_verbytes = xpub_verbytes
        self.xprv_verbytes = xprv_verbytes

    def bits_to_difficulty(self, bits):
        return Bitcoin.max_target / bits_to_target(bits)

    def deserialized_header(self, raw, height):
        '''Returns a deserialized header object.'''
        return Header(*unpack_header(raw), self.header_hash(raw), raw, height)

    def header_hash(self, raw_header):
        return double_sha256(raw_header)

    def header_prev_hash(self, raw_header):
        return raw_header[4:36]

    def header_timestamp(self, raw_header):
        timestamp, = unpack_le_uint32(raw_header[68:72])
        return timestamp

    def header_bits(self, raw_header):
        bits, = unpack_le_uint32(raw_header[72:76])
        return bits

    def header_work(self, raw_header):
        return bits_to_work(self.header_bits(raw_header))

    @classmethod
    def from_WIF_byte(cls, WIF_byte):
        '''Return the coin using the given WIF byte.'''
        for coin in all_coins:
            if WIF_byte == coin.WIF_byte:
                return coin
        raise ValueError(f'invalid WIF byte {WIF_byte}')

    @classmethod
    def lookup_xver_bytes(cls, xver_bytes):
        '''Returns a (coin, is_public_key) pair.'''
        for coin in all_coins:
            if xver_bytes == coin.xpub_verbytes:
                return coin, True
            if xver_bytes == coin.xprv_verbytes:
                return coin, False
        raise ValueError(f'invalid xver_bytes {xver_bytes}')


Bitcoin = Coin(
    'Bitcoin mainnet',
    '0100000000000000000000000000000000000000000000000000000000000000000000003ba3edfd'
    '7a7b12b27ac72c3e67768f617fc81bc3888a51323a9fb8aa4b1e5e4a29ab5f49ffff001d1dac2b7c',
    required_bits_mainnet,
    0x00,
    0x80,
    bytes.fromhex("0488b21e"),
    bytes.fromhex("0488ade4"),
)


BitcoinTestnet = Coin(
    'Bitcoin testnet',
    '0100000000000000000000000000000000000000000000000000000000000000000000003ba3edfd'
    '7a7b12b27ac72c3e67768f617fc81bc3888a51323a9fb8aa4b1e5e4adae5494dffff001d1aa4ae18',
    required_bits_testnet,
    0x6f,
    0xef,
    bytes.fromhex("043587cf"),
    bytes.fromhex("04358394"),
)

BitcoinScalingTestnet = Coin(
    'Bitcoin scaling testnet',
    '0100000000000000000000000000000000000000000000000000000000000000000000003ba3edfd'
    '7a7b12b27ac72c3e67768f617fc81bc3888a51323a9fb8aa4b1e5e4adae5494dffff001d1aa4ae18',
    required_bits_scaling_testnet,
    0x6f,
    0xef,
    bytes.fromhex("043587cf"),
    bytes.fromhex("04358394"),
)

# rt12 -- Scaling testnet has same settings as regular testnet, so will cause conflicts.
all_coins = (Bitcoin, BitcoinTestnet, BitcoinScalingTestnet)



# Copyright (c) 2019, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''Public and Private keys of various kinds.'''

__all__ = (
    'PrivateKey', 'PublicKey', 'CURVE_ORDER',
    'KeyException', 'InvalidSignatureError', 'DecryptionError',
    'der_signature_to_compact', 'compact_signature_to_der',
)

from base64 import b64decode, b64encode
from binascii import Error as binascii_Error
from os import urandom

from electrumsv_secp256k1 import ffi, lib, create_context

from .aes import aes_encrypt_with_iv, aes_decrypt_with_iv
from .base58 import base58_encode_check, base58_decode_check, is_minikey
from .coin import Bitcoin, Coin
from .hashes import sha256, sha512, double_sha256, hash160, hmac_digest, _sha256
from .misc import be_bytes_to_int, int_to_be_bytes
from .packing import pack_byte, pack_varbytes
from .script import Script
from .util import cachedproperty


KEY_SIZE = 32
CURVE_ORDER = 0xfffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141
CONTEXT = create_context()
CDATA_SIG_LENGTH = 64
MAX_SIG_LENGTH = 72
EC_COMPRESSED = lib.SECP256K1_EC_COMPRESSED
EC_UNCOMPRESSED = lib.SECP256K1_EC_UNCOMPRESSED
SIGNED_MESSAGE_PREFIX = pack_varbytes('Bitcoin Signed Message:\n'.encode())


class KeyException(Exception):
    pass


class InvalidSignatureError(KeyException):
    pass


class DecryptionError(KeyException):
    pass


def _to_32_bytes(value):
    if not isinstance(value, bytes):
        raise TypeError('value must have type bytes')
    if len(value) != 32:
        raise ValueError('value must be 32 bytes')
    return bytes(value)


def _message_hash(message, hasher):
    msg_hash = hasher(message) if hasher is not None else message
    if len(msg_hash) != 32:
        raise ValueError('hashed message must be 32 bytes')
    return msg_hash


def _serialize_der(signature):
    '''Return a DER-serialized signature; bytes of length at most 72.'''
    size = ffi.new('size_t *', MAX_SIG_LENGTH)
    data = ffi.new(f'unsigned char [{MAX_SIG_LENGTH}]')
    result = lib.secp256k1_ecdsa_signature_serialize_der(CONTEXT, data, size, signature)
    # Failure should never happen - MAX_SIG_LENGTH bytes is always enough
    assert result
    return bytes(ffi.buffer(data, size[0]))


def _serialize_recoverable(recover_sig, context):
    '''Return a 65-byte compact serialized signature.'''
    output = ffi.new(f'unsigned char [{CDATA_SIG_LENGTH}]')
    recid = ffi.new('int *')

    lib.secp256k1_ecdsa_recoverable_signature_serialize_compact(
        context, output, recid, recover_sig)

    # recid is 0, 1, 2 or 3.
    return bytes(ffi.buffer(output, CDATA_SIG_LENGTH)) + bytes([recid[0]])


def _deserialize_recoverable_sig(recover_sig_bytes):
    if len(recover_sig_bytes) != 65:
        raise InvalidSignatureError('invalid recoverable signature')
    recoverable_sig = ffi.new('secp256k1_ecdsa_recoverable_signature *')
    recid = recover_sig_bytes[-1]
    if not 0 <= recid <= 3:
        raise InvalidSignatureError('invalid recoverable signature')
    if not lib.secp256k1_ecdsa_recoverable_signature_parse_compact(
            CONTEXT, recoverable_sig, recover_sig_bytes, recid):
        raise InvalidSignatureError('invalid recoverable signature')
    return recoverable_sig


def _message_sig_to_recoverable_sig(message_sig):
    '''Return a recoverable signature from a message signature.'''
    if isinstance(message_sig, str):
        try:
            message_sig = b64decode(message_sig, validate=True)
        except binascii_Error:
            raise InvalidSignatureError('invalid base64 encoding of message signature')
    if not isinstance(message_sig, bytes) or len(message_sig) != 65:
        raise InvalidSignatureError('message signature must be 65 bytes')
    if not 27 <= message_sig[0] < 35:
        raise InvalidSignatureError('invalid message signature format')
    return message_sig[1:] + pack_byte((message_sig[0] - 27) & 3)


def _recoverable_sig_to_message_sig(recoverable_sig, compressed):
    leading_byte = 27 + (4 if compressed else 0) + recoverable_sig[-1]
    return pack_byte(leading_byte) + recoverable_sig[:64]


def _recoverable_sig_to_cdata_sig(recoverable_sig):
    cdata_sig = ffi.new('secp256k1_ecdsa_signature *')
    rec_sig = _deserialize_recoverable_sig(recoverable_sig)
    # Always succeeds
    lib.secp256k1_ecdsa_recoverable_signature_convert(CONTEXT, cdata_sig, rec_sig)
    return cdata_sig


def _der_sig_to_cdata_sig(der_sig):
    cdata_sig = ffi.new('secp256k1_ecdsa_signature *')
    if not lib.secp256k1_ecdsa_signature_parse_der(CONTEXT, cdata_sig, der_sig, len(der_sig)):
        raise InvalidSignatureError('invalid DER-encoded signature')
    return cdata_sig


def _normalize_message(message):
    if isinstance(message, str):
        message = message.encode()
    return SIGNED_MESSAGE_PREFIX + pack_varbytes(message)


def der_signature_to_compact(der_sig):
    '''Returns 64 bytes representing r and s as concatenated 32-byte big-endian numbers.'''
    cdata_sig = _der_sig_to_cdata_sig(der_sig)
    compact_sig = ffi.new('unsigned char [64]')
    lib.secp256k1_ecdsa_signature_serialize_compact(CONTEXT, compact_sig, cdata_sig)
    return bytes(ffi.buffer(compact_sig))


def compact_signature_to_der(compact_sig):
    if not (isinstance(compact_sig, bytes) and len(compact_sig) == 64):
        raise InvalidSignatureError('compact signature must be 64 bytes')
    cdata_sig = ffi.new('secp256k1_ecdsa_signature *')
    lib.secp256k1_ecdsa_signature_parse_compact(CONTEXT, cdata_sig, compact_sig)
    return _serialize_der(cdata_sig)


class PrivateKey:

    def __init__(self, secret, compressed=True, coin=None):
        '''Construct a PrivateKey from 32 big-endian bytes.

        compressed is passed on to the PublicKey constructor to indicate whether the
        public key should return an compressed serialization.

        A private key exists independently of any coin.  However it is useful to have a
        default coin for certain methods, such as to_WIF().  A PrivateKey may be created
        with an implicit coin (e.g. if created from WIF), so in such cases we remember it.
        If there is no implicit coin and client code does specify one, Bitcoin is used.
        '''
        if isinstance(secret, bytes):
            if len(secret) != KEY_SIZE:
                raise ValueError('private key must be 32 bytes')
            if not lib.secp256k1_ec_seckey_verify(CONTEXT, secret):
                raise ValueError('private key out of range')
            self._secret = secret
        elif repr(secret).startswith("<cdata 'unsigned char[32]'"):
            self._secret = bytes(ffi.buffer(secret))
        else:
            raise TypeError('private key must be bytes')
        self._compressed = compressed
        self._coin = coin or Bitcoin

    def _secp256k1_public_key(self):
        '''Construct a wrapped secp256k1 PublicKey.'''
        public_key = ffi.new('secp256k1_pubkey *')
        created = lib.secp256k1_ec_pubkey_create(CONTEXT, public_key, self._secret)
        # Only possible if client code has mucked with the private key's internals
        if not created:
            raise RuntimeError('invalid private key')
        return public_key

    # Public methods

    def __eq__(self, other):
        '''Return True if this PrivateKey is equal to another.'''
        return self._secret == other._secret

    def __str__(self):
        '''Return a hash of the private key, out of an abundance of caution.
        To get a real string call to_hex() explicitly.'''
        return sha256(self._secret).hex()

    def coin(self):
        '''Returns an implied coin if there is one, otherwise Bitcoin.'''
        return self._coin

    def is_compressed(self):
        '''Return true if the public key serializes to 33 bytes.'''
        return self._compressed

    @cachedproperty
    def public_key(self):
        '''Return a PublicKey corresponding to this private key.'''
        return PublicKey(self._secp256k1_public_key(), self._compressed, self._coin)

    def to_int(self):
        '''Return the private key's representation as an integer.'''
        return be_bytes_to_int(self._secret)

    def to_hex(self):
        '''Return the private key's representation as a hexidecimal string.'''
        return self._secret.hex()

    def to_bytes(self):
        '''Return the private key's representation as bytes (32 bytes, big-endian).'''
        return self._secret

    @classmethod
    def from_int(cls, value):
        '''Contruct a PrivateKey from an unsigned integer.'''
        return cls(int_to_be_bytes(value, 32))

    @classmethod
    def from_arbitrary_bytes(cls, secret):
        return cls.from_int(be_bytes_to_int(secret) % CURVE_ORDER)

    @classmethod
    def from_hex(cls, hex_str):
        '''Contruct a PrivateKey from a hexadecimal string of 64 characters.

        There is no automatic padding.'''
        return cls(bytes.fromhex(hex_str))

    @classmethod
    def from_minikey(cls, minikey):
        '''Construct a PrivateKey from its Minikey encoding (used in Casascius coins).

        Minikeys used uncompressed public keys.'''
        if not is_minikey(minikey):
            raise ValueError('invalid minikey')
        return cls(sha256(minikey.encode()), False)

    @classmethod
    def from_random(cls, *, source=urandom):
        '''Return a random, valid PrivateKey.'''
        while True:
            try:
                return cls(source(32), True)
            except ValueError:
                pass

    @classmethod
    def from_WIF(cls, txt):
        '''Construct a PriveKey from WIF text.'''
        raw = base58_decode_check(txt)
        if len(raw) == 33 or len(raw) == 34 and raw[-1] == 0x01:
            return cls(raw[1:33], len(raw) == 34, Coin.from_WIF_byte(raw[0]))
        raise ValueError('invalid WIF private key')

    def to_WIF(self, *, compressed=None, coin=None):
        '''Return the WIF form of the private key for the given coin.

        Set compressed to True to indicate the corresponding public key should be the
        compressed form.
        '''
        coin = coin or self._coin
        payload = pack_byte(coin.WIF_byte) + self._secret
        if (self._compressed if compressed is None else compressed):
            payload += pack_byte(0x01)
        return base58_encode_check(payload)

    def add(self, value):
        '''Return a new PrivateKey instance adding value to our secret.'''
        secret = ffi.new('unsigned char [32]', self._secret)
        if not lib.secp256k1_ec_privkey_tweak_add(CONTEXT, secret, _to_32_bytes(value)):
            raise ValueError('value or result out of range')
        return PrivateKey(secret, self._compressed, self._coin)

    def multiply(self, value):
        '''Return a new PrivateKey instance multiplying value by our secret.'''
        secret = ffi.new('unsigned char [32]', self._secret)
        if not lib.secp256k1_ec_privkey_tweak_mul(CONTEXT, secret, _to_32_bytes(value)):
            raise ValueError('value or result out of range')
        return PrivateKey(secret, self._compressed, self._coin)

    def sign(self, message, hasher=sha256):
        '''Sign a message (more correctly its hash) and return a DER-encoded signature.

        If the message is already hashed, set hasher to None.
        '''
        msg_hash = _message_hash(message, hasher)
        signature = ffi.new('secp256k1_ecdsa_signature *')
        if not lib.secp256k1_ecdsa_sign(CONTEXT, signature, msg_hash, self._secret,
                                        ffi.NULL, ffi.NULL):
            raise ValueError('invalid private key')
        return _serialize_der(signature)

    def sign_recoverable(self, message, hasher=sha256):
        '''Sign a message (more correctly its hash) and return a byte of metadata so that the
        public key can be immediately recovered from the signature.

        If the message is already hashed, set hasher to None.
        '''
        msg_hash = _message_hash(message, hasher)
        signature = ffi.new('secp256k1_ecdsa_recoverable_signature *')
        if not lib.secp256k1_ecdsa_sign_recoverable(CONTEXT, signature, msg_hash, self._secret,
                                                    ffi.NULL, ffi.NULL):
            raise ValueError('invalid private key')
        return _serialize_recoverable(signature, CONTEXT)

    def sign_message(self, message, hasher=double_sha256):
        '''Sign a message compatibly with bitcoind (and ElectrumSV).

        message:     the message as bytes before being prefixed with SIGNED_MESSAGE_PREFIX.
                     If a string, it is first UTF-8 encoded to bytes before prefixing.
        hasher:      used to hash the message to 32 bytes.  Cannot be None as message cannot be
                     a hash.

        Returns a 65-byte signature.
        '''
        if hasher is None:
            raise ValueError('hasher cannot be None')
        message = _normalize_message(message)
        recoverable_sig = self.sign_recoverable(message, hasher)
        # Compressed appears to be legacy and the signature is valid whether True or
        # False; in any case only the first byte (recid) of the signature changes.
        return _recoverable_sig_to_message_sig(recoverable_sig, compressed=self._compressed)

    def sign_message_to_base64(self, message, hasher=double_sha256):
        '''As for sign_message, but return the result as a base64 ASCII string.'''
        return b64encode(self.sign_message(message, hasher)).decode()

    def shared_secret(self, public_key, message, hasher=sha256):
        '''Return a shared secret (as a public key) given their public key and a message.

        The deterministic key is formed by hashing the message; alternatively it can be
        passed in directly if hasher is None.
        '''
        deterministic_key = _message_hash(message, hasher)
        private_key2 = self.add(deterministic_key)
        public_key2 = public_key.add(deterministic_key)
        return public_key2.multiply(private_key2._secret)

    def ecdh_shared_secret(self, public_key):
        '''Return an Elliptic Curve Diffie-Helman shared secret (as a public key) given their
        public key.

        This is a degenerate form of shared_secret() where the deterministic key is zero.
        '''
        return public_key.multiply(self._secret)

    def decrypt_message(self, message, magic=b'BIE1'):
        '''Decrypt a message encrypted with PublicKey.encrypt_message().'''
        if isinstance(message, str):
            try:
                message = b64decode(message, validate=True)
            except binascii_Error:
                raise DecryptionError('invalid base64 encoding of encrypted message')
        mlen = len(magic)
        if len(message) < 81 + mlen:
            raise DecryptionError('message too short')
        if not message.startswith(magic):
            raise DecryptionError('bad magic')

        try:
            ephemeral_pubkey = PublicKey.from_bytes(message[mlen: mlen + 33])
        except ValueError as e:
            raise DecryptionError(f'invalid ephemeral public key: {e}') from None

        ciphertext = message[mlen + 33:-32]
        hmac = message[-32:]
        key = sha512(self.ecdh_shared_secret(ephemeral_pubkey).to_bytes())
        iv, key_e, key_m = key[0:16], key[16:32], key[32:]

        if hmac_digest(key_m, message[:-32], _sha256) != hmac:
            raise DecryptionError('bad HMAC')

        try:
            return aes_decrypt_with_iv(key_e, iv, ciphertext)
        except Exception as e:  # aes library can raise Exception, unfortunately
            raise DecryptionError(f'{e}') from None


class PublicKey:

    def __init__(self, public_key, compressed, coin=None):
        '''Construct a PublicKey.

        This function is not intended to be called directly by user code; use instead one
        of the "from_" class methods or using a PrivateKey's 'public_key' property.

        compressed is True if to_bytes() serializations yield public keys of 33 bytes.

        A public key exists independently of any coin.  However it is useful to have a
        default coin for certain methods, such as __str__().
        If there is no implicit coin and client code does specify one, Bitcoin is used.
        '''
        if not repr(public_key).startswith("<cdata 'secp256k1_pubkey *'"):
            raise TypeError('PublicKey constructor requires a secp256k1_pubkey')
        self._public_key = public_key
        self._compressed = compressed
        self._coin = coin or Bitcoin

    # Public methods

    def __eq__(self, other):
        '''Return True if this PublicKey is equal to another.'''
        return self.to_bytes(compressed=True) == other.to_bytes(compressed=True)

    def __str__(self):
        return self.to_hex()

    def coin(self):
        '''Returns an implied coin if there is one, otherwise Bitcoin.'''
        return self._coin

    def is_compressed(self):
        '''Return true if it serializes to 33 bytes.'''
        return self._compressed

    def to_bytes(self, *, compressed=None):
        '''Serialize a PublicKey to bytes.'''
        if (self._compressed if compressed is None else compressed):
            length, flag = 33, EC_COMPRESSED
        else:
            length, flag = 65, EC_UNCOMPRESSED
        result = ffi.new(f'unsigned char [{length}]')
        rlength = ffi.new('size_t *', length)

        lib.secp256k1_ec_pubkey_serialize(CONTEXT, result, rlength, self._public_key, flag)
        return bytes(ffi.buffer(result, length))

    @classmethod
    def from_bytes(cls, data):
        '''Construct a PublicKey from its serialized bytes.

        data should be bytes of length 33 or 65.'''
        public_key = ffi.new('secp256k1_pubkey *')
        if not lib.secp256k1_ec_pubkey_parse(CONTEXT, public_key, data, len(data)):
            raise ValueError('invalid public key')
        return cls(public_key, len(data) == 33)

    @classmethod
    def from_recoverable_signature(cls, recoverable_sig, message, hasher=sha256):
        '''Constuct a PublicKey from a recoverable signature and message (hash) that was
        signed.'''
        msg_hash = _message_hash(message, hasher)
        recoverable_sig = _deserialize_recoverable_sig(recoverable_sig)
        public_key = ffi.new('secp256k1_pubkey *')
        if not lib.secp256k1_ecdsa_recover(CONTEXT, public_key, recoverable_sig, msg_hash):
            raise InvalidSignatureError('invalid recoverable signature')
        return cls(public_key, True)

    @classmethod
    def from_signed_message(cls, message_sig, message, hasher=double_sha256):
        '''Contruct a PublicKey from a message and its signature.

        message_sig: 65 bytes; if a string assumed base64-encoded.
        message:     the message as bytes before being prefixed with SIGNED_MESSAGE_PREFIX.
                     If a string, it is first UTF-8 encoded to bytes before prefixing.
        hasher:      used to hash the message to 32 bytes.  Cannot be None as message cannot be
                     a hash.
        '''
        if hasher is None:
            raise ValueError('hasher cannot be None')
        message = _normalize_message(message)
        recoverable_sig = _message_sig_to_recoverable_sig(message_sig)
        return cls.from_recoverable_signature(recoverable_sig, message, hasher)

    def to_hex(self, *, compressed=None):
        '''Convert a PublicKey to a hexadecimal string.'''
        return self.to_bytes(compressed=compressed).hex()

    @classmethod
    def from_hex(cls, hex_str):
        '''Construct a PublicKey from a hexadecimal string.'''
        return cls.from_bytes(bytes.fromhex(hex_str))

    def to_point(self):
        '''Return the PublicKey as an (x, y) point on the curve.'''
        data = self.to_bytes(compressed=False)
        x = be_bytes_to_int(data[1:33])
        y = be_bytes_to_int(data[33:])
        return x, y

    @classmethod
    def from_point(cls, x, y):
        '''Construct a PublicKey from a (x, y) point on the curve.'''
        x_bytes = int_to_be_bytes(x, 32)
        y_bytes = int_to_be_bytes(y, 32)
        return cls.from_bytes(b''.join((b'\x04', x_bytes, y_bytes)))

    def to_address(self, *, compressed=None, coin=None):
        '''Return the public key as a bitcoin P2PKH address.'''
        coin = coin or self.coin()
        data = self.to_bytes(compressed=compressed)
        return base58_encode_check(pack_byte(coin.P2PKH_verbyte) + hash160(data))

    def add(self, value):
        '''Return a new PublicKey instance formed by adding value*G to this one.

        Preserves compressed / uncompressed serialization.'''
        public_key = ffi.new('secp256k1_pubkey *', self._public_key[0])
        if not lib.secp256k1_ec_pubkey_tweak_add(CONTEXT, public_key, _to_32_bytes(value)):
            raise ValueError('value or result out of range')
        return PublicKey(public_key, self._compressed, self._coin)

    def multiply(self, value):
        '''Return a new PublicKey instance formed by multiplying this one by value (i.e. adding
        it to itself value times).

        Preserves compressed / uncompressed serialization.
        '''
        public_key = ffi.new('secp256k1_pubkey *', self._public_key[0])
        if not lib.secp256k1_ec_pubkey_tweak_mul(CONTEXT, public_key, _to_32_bytes(value)):
            raise ValueError('value or result out of range')
        return PublicKey(public_key, self._compressed, self._coin)

    @classmethod
    def combine_keys(cls, public_keys):
        '''Return a new PublicKey equal to the sum of the given PublicKeys.

        The result takes its default compressed and coin attributes from the first public key.
        '''
        lib_keys = [pk._public_key for pk in public_keys]
        if not lib_keys:
            raise ValueError('no public keys to combine')
        public_key = ffi.new('secp256k1_pubkey *')
        if not lib.secp256k1_ec_pubkey_combine(CONTEXT, public_key, lib_keys, len(lib_keys)):
            raise ValueError('the sum of the public keys is invalid')
        return cls(public_key, public_keys[0]._compressed, public_keys[0]._coin)

    def verify_message(self, message_sig, message, hasher=double_sha256):
        '''Verify a message signed with bitcoind (and ElectrumSV).

        message_sig: 65 bytes; if a string assumed base64-encoded.
        message:     the message as bytes before being prefixed with SIGNED_MESSAGE_PREFIX.
                     If a string, it is first UTF-8 encoded to bytes before prefixing.
        hasher:      used to hash the message to 32 bytes.  Cannot be None as message cannot be
                     a hash.
        '''
        if hasher is None:
            raise ValueError('hasher cannot be None')
        message = _normalize_message(message)
        recoverable_sig = _message_sig_to_recoverable_sig(message_sig)
        return self.verify_recoverable_signature(recoverable_sig, message, hasher)

    @classmethod
    def verify_message_and_address(cls, message_sig, message, address, *,
                                   hasher=double_sha256, coin=None):
        '''As for verify_message, but also test it was signed by a private key of the given
        address.
        '''
        try:
            public_key = cls.from_signed_message(message_sig, message, hasher)
        except InvalidSignatureError:
            return False
        return (public_key.verify_message(message_sig, message, hasher) and
                any(address == public_key.to_address(compressed=compressed, coin=coin)
                    for compressed in (True, False)))

    def verify_der_signature(self, der_sig, message, hasher=sha256):
        '''Verify a der-encoded signature.  Return True if good otherwise False.'''
        cdata_sig = _der_sig_to_cdata_sig(der_sig)
        msg_hash = _message_hash(message, hasher)
        return bool(lib.secp256k1_ecdsa_verify(CONTEXT, cdata_sig, msg_hash, self._public_key))

    def verify_recoverable_signature(self, recoverable_sig, message, hasher=sha256):
        '''Verify a recoverable signature.  Return True if good otherwise False.'''
        cdata_sig = _recoverable_sig_to_cdata_sig(recoverable_sig)
        msg_hash = _message_hash(message, hasher)
        return bool(lib.secp256k1_ecdsa_verify(CONTEXT, cdata_sig, msg_hash, self._public_key))

    def encrypt_message(self, message, magic=b'BIE1'):
        '''Encrypt a message using ECIES.  The message can be bytes or a string.

        String are converted to UTF-8 encoded bytes.  The result has type bytes.
        '''
        if isinstance(message, str):
            message = message.encode()
        ephemeral_key = PrivateKey.from_random()
        key = sha512(ephemeral_key.ecdh_shared_secret(self).to_bytes(compressed=True))
        iv, key_e, key_m = key[0:16], key[16:32], key[32:]
        ciphertext = aes_encrypt_with_iv(key_e, iv, message)
        ephemeral_pubkey = ephemeral_key.public_key.to_bytes()
        encrypted_data = b''.join((magic, ephemeral_pubkey, ciphertext))
        return encrypted_data + hmac_digest(key_m, encrypted_data, _sha256)

    def encrypt_message_to_base64(self, message, magic=b'BIE1'):
        '''As for encrypt_message, but return the result as a base64 ASCII string.'''
        return b64encode(self.encrypt_message(message, magic)).decode()

    def P2PK_script(self, *, compressed=None):
        return Script.P2PK_script(self.to_bytes(compressed=compressed))

    def P2PKH_script(self, *, compressed=None):
        return Script.P2PKH_script(hash160(self.to_bytes(compressed=compressed)))



# Copyright (c) 2016-2018, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''Cryptographic hash functions and related classes.'''

__all__ = (
    'sha1', 'sha256', 'sha512', 'double_sha256',
    'ripemd160', 'hash160',
    'hmac_digest', 'hmac_sha512_halves',
    'hash_to_hex_str', 'hex_str_to_hash', 'hash_to_value',
)

import hashlib
import hmac


_sha1 = hashlib.sha1
_sha256 = hashlib.sha256
_sha512 = hashlib.sha512
_new_hash = hashlib.new
bytes_fromhex = bytes.fromhex


def sha1(x):
    '''Simple wrapper of hashlib sha1.'''
    return _sha1(x).digest()


def sha256(x):
    '''Simple wrapper of hashlib sha256.'''
    return _sha256(x).digest()


def sha512(x):
    '''Simple wrapper of hashlib sha256.'''
    return _sha512(x).digest()


def ripemd160(x):
    '''Simple wrapper of hashlib ripemd160.'''
    h = _new_hash('ripemd160')
    h.update(x)
    return h.digest()


def double_sha256(x):
    '''SHA-256 of SHA-256, as used extensively in bitcoin.'''
    return sha256(sha256(x))


def hash160(x):
    '''RIPEMD-160 of SHA-256.

    Used to make bitcoin addresses from pubkeys.'''
    return ripemd160(sha256(x))


def hash_to_hex_str(x):
    '''Convert a little-endian binary hash to displayed hex string.

    Display form of a binary hash is reversed and converted to hex.
    '''
    return bytes(reversed(x)).hex()


def hash_to_value(x):
    '''Convert a little-endian binary hash to an integer value (to e.g. compare to a proof of
    work target).
    '''
    return int.from_bytes(x, 'little')


def hex_str_to_hash(x):
    '''Convert a displayed hex string to a binary hash.'''
    return bytes(reversed(bytes_fromhex(x)))


if hasattr(hmac, 'digest'):
    # Python 3.7+
    hmac_digest = hmac.digest   # pylint: disable=no-member
else:
    def hmac_digest(key, msg, digest):
        return hmac.new(key, msg, digest).digest()


def hmac_sha512_halves(key, msg):
    hmacd = hmac_digest(key, msg, _sha512)
    return hmacd[:32], hmacd[32:]



# Copyright (c) 2017-2019, Neil Booth
#
# All rights reserved.
#
# See the file "LICENCE" for information about the copyright
# and warranty status of this software.

'''Base58 encoding.'''

__all__ = (
    'base58_decode', 'base58_encode', 'base58_decode_check', 'base58_encode_check',
    'Base58Error', 'is_minikey',
)

from .hashes import double_sha256, sha256
from .misc import int_to_be_bytes, be_bytes_to_int


base58_chars = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'
base58_cmap = {c: n for n, c in enumerate(base58_chars)}


class Base58Error(Exception):
    '''Exception used for Base58 errors.'''


def base58_decode(txt):
    """Decodes txt into a big-endian bytearray."""
    if not isinstance(txt, str):
        raise TypeError('a string is required')

    if not txt:
        raise Base58Error('string cannot be empty')

    cvalue = base58_cmap.get
    value = 0
    try:
        for c in txt:
            value = value * 58 + cvalue(c)
    except TypeError:
        raise Base58Error(f'invalid base 58 character "{c}"') from None

    result = int_to_be_bytes(value)

    # Prepend leading zero bytes if necessary
    count = 0
    for c in txt:
        if c != '1':
            break
        count += 1
    if count:
        result = bytes(count) + result

    return result


def base58_encode(be_bytes):
    '''Convert a big-endian bytearray into a base58 string.'''
    value = be_bytes_to_int(be_bytes)

    txt = ''
    while value:
        value, mod = divmod(value, 58)
        txt += base58_chars[mod]

    for byte in be_bytes:
        if byte != 0:
            break
        txt += '1'

    return txt[::-1]


def base58_decode_check(txt):
    '''Decode a Base58Check-encoded string to a payload.  The version prefixes it.'''
    be_bytes = base58_decode(txt)
    result, check = be_bytes[:-4], be_bytes[-4:]
    if check != double_sha256(result)[:4]:
        raise Base58Error(f'invalid base 58 checksum for {txt}')
    return result


def base58_encode_check(payload):
    '''Encode a payload bytearray (which includes the version byte(s)) into a Base58Check
    string.
    '''
    be_bytes = payload + double_sha256(payload)[:4]
    return base58_encode(be_bytes)


def is_minikey(text):
    # Minikeys are 22 or 30 characters.  A valid minikey must begin with an 'S', be in
    # base58, and when suffixed with '?' have its SHA256 hash begin with a zero byte.
    # They are widely used in Casascius physical bitcoins, where the address corresponded
    # to an uncompressed public key.
    return (len(text) in (22, 30)
            and text[0] == 'S'
            and all(c in base58_cmap for c in text)
            and sha256((text + '?').encode())[0] == 0x00)



# Copyright (c) 2018, 2019, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

__all__ = (
    'Chain', 'CheckPoint', 'Headers',
    'ChainException', 'MissingHeader', 'IncorrectBits', 'InsufficientPoW',
)

import array
from collections import namedtuple
import logging
import math
from struct import Struct

from bitcoinx.hashes import hash_to_hex_str
from bitcoinx.packing import pack_le_uint32, unpack_le_uint32
from bitcoinx.util import map_file


empty_header = bytes(80)
logger = logging.getLogger('chain')


# The prev_work of a checkpoint is the cumulative work prior to the checkpoint and does
# not including the work of the checkpoint itself
CheckPoint = namedtuple('CheckPoint', 'raw_header height prev_work')


class ChainException(Exception):
    pass


class MissingHeader(ChainException):
    pass


class IncorrectBits(ChainException):

    def __init__(self, header, required_bits):
        super().__init__(header, required_bits)
        self.header = header
        self.required_bits = required_bits

    def __str__(self):
        return f'header f{self.header} requires bits 0x{self.required_bits}'


class InsufficientPoW(ChainException):
    def __init__(self, header):
        super().__init__(header)
        self.header = header

    def __str__(self):
        return (f'header f{self.header} hash value f{self.header.hash_value()} exceeds '
                f'its target {self.header.target()}')


class _BadHeadersFile(Exception):
    pass


class Chain(object):
    '''A dumb object representing a chain of headers back to the genesis block (implemented
    through parent chains).

    Public attributes:
        parent        the parent chain this one forks from, can be None
        first_height  the first height not in common with the parent (0 for the base chain)
        height        the height of the chain
        tip           the deserialized Header object of the chain tip
        work          cumulative chain work to the tip
    '''

    def __init__(self, parent, tip, tip_header_index, prev_work):
        '''common_height is the greatest height common to the chain and its parent.'''
        self.parent = parent
        self.tip = tip
        self.work = prev_work + tip.work()
        self.first_height = tip.height
        self._header_indices = array.array('I')
        self._header_indices.append(tip_header_index)

    @classmethod
    def from_checkpoint(cls, coin, checkpoint):
        tip = coin.deserialized_header(checkpoint.raw_header, checkpoint.height)
        return cls(None, tip, tip.height, checkpoint.prev_work)

    def append(self, header, header_index):
        '''Append a header to the chain along with its index in storage.'''
        self.tip = header
        self._header_indices.append(header_index)
        self.work += header.work()

    def parent_heights(self):
        '''Returns a map {chain: height}.  The keys are the chain and every parent chain,
        recursively.  The height is the last height on that chain for parent chains, and
        the chain height for ourself.
        '''
        result = {self: self.height}
        chain = self
        while chain.parent:
            result[chain.parent] = chain.first_height - 1
            chain = chain.parent
        return result

    def common_chain_and_height(self, other_chain):
        '''Returns a pair (chain, height).  The height is the greatest height common between this
        chain and another chain back to the genesis block, and chain is the chain of that
        height.
        '''
        other_heights = other_chain.parent_heights()
        our_heights = self.parent_heights()
        result = (None, -1)
        for chain in our_heights:
            if chain in other_heights:
                common_height = min(other_heights[chain], our_heights[chain])
                if common_height > result[1]:
                    result = (chain, common_height)
        return result

    def header_index(self, height):
        '''Return the index of the header in storage.'''
        if height >= self.first_height:
            try:
                return self._header_indices[height - self.first_height]
            except IndexError:
                pass
        elif self.parent:
            return self.parent.header_index(height)
        elif height >= 0:
            return height
        raise MissingHeader(f'no header at height {height}')

    def log2_work(self):
        return math.log(self.work, 2)

    @property
    def height(self):
        return self.first_height + len(self._header_indices) - 1

    def desc(self):
        return f'tip={self.tip} log2_work={round(self.log2_work(), 8)}'


class _HeaderStorage(object):
    '''Implementation of raw header storage for flat files.

    Block headers are looked up by index, which in general is not equal to its height.

    If a block header is read that has not been set, MissingHeader is raised.

    Block headers are 80 bytes; the caller is responsible for their validation.

    Flat files are stored as a reserved area, followed by the headers consecutively.
    The reserved area has the following format:
       a) reserved area size (little endian uint16)
       b) version number (little endian uint16)
       c) block header count (little endian uint32)
    '''
    struct_reserved = Struct('<HHI')

    def __init__(self, filename):
        '''Create an object representing flat file header storage in filename.'''
        self.filename = filename
        self.mmap = None
        self.reserved_size = self.struct_reserved.size
        self.header_count = 0

    def _offset(self, key):
        return self.reserved_size + key * 80

    def _create_file(self, checkpoint):
        s = self.struct_reserved
        self.reserved_size = s.size
        with open(self.filename, 'wb') as f:
            f.write(s.pack(s.size, 0, checkpoint.height + 1))
            f.seek(self._offset(checkpoint.height))
            f.write(checkpoint.raw_header)

    def _open_file(self, checkpoint):
        logger.debug(f'opening headers file {self.filename}')
        s = self.struct_reserved
        self.mmap = map_file(self.filename)
        try:
            if len(self.mmap) >= s.size:
                self.reserved_size, version, self.header_count = s.unpack(self.mmap[:s.size])
                # Note self[checkpoint.height] might raise MissingHeader
                if version == 0 and self[checkpoint.height] == checkpoint.raw_header:
                    return
            raise _BadHeadersFile(f'invalid headers file {self.filename}')
        except Exception:
            self.mmap.close()
            self.mmap = None
            raise

    def open_or_create(self, checkpoint):
        try:
            self._open_file(checkpoint)
            return
        except FileNotFoundError:
            logger.debug(f'{self.filename} not found, creating it')
        except (_BadHeadersFile, MissingHeader):
            logger.debug(f're-creating headers file {self.filename}')
        self._create_file(checkpoint)
        self._open_file(checkpoint)

    def _set_count(self, count):
        self.mmap[4:8] = pack_le_uint32(count)

    def _set_raw_header(self, index, raw_header):
        if not isinstance(raw_header, (bytes, bytearray)) or len(raw_header) != 80:
            raise TypeError('raw header must be binary of length 80')
        # Grow if needed
        mmap = self.mmap
        start = self._offset(index)
        if start >= len(mmap):
            mmap.close()
            mmap = self.mmap = map_file(self.filename, self._offset(index + 5_000))
        if index >= len(self):
            self._set_count(index + 1)
        self.mmap[start: start + 80] = raw_header

    def __getitem__(self, key):
        def header(index):
            start = self._offset(index)
            result = self.mmap[start: start + 80]
            if not result or result == empty_header:
                raise MissingHeader(f'no header at index {index}')
            return result

        if isinstance(key, int):
            if key < 0:
                key += len(self)
            return header(key)
        elif isinstance(key, slice):
            return [header(index) for index in range(*key.indices(len(self)))]
        raise TypeError(f'key {key} should be an integer')

    def __setitem__(self, key, raw_header):
        if isinstance(key, int):
            self._set_raw_header(key, raw_header)
        else:
            raise TypeError(f'key {key} should be an integer')

    def __len__(self):
        count, = unpack_le_uint32(self.mmap[4:8])
        return count

    def append(self, raw_header):
        header_index = len(self)
        self._set_raw_header(header_index, raw_header)
        return header_index

    def close(self):
        self.mmap.close()

    def flush(self):
        return self.mmap.flush()


class Headers(object):
    '''A collection of block headers, including a checkpoint header, arranged into chains.
    Each header header belongs to precisely one chain.  Each chain has a parent chain
    which it forked from, except one chain whose parent is None.

    Headers before the checkpoint can be set and no validation is performed; the caller is
    presumed to have done any necessary validation.  Headers after the checkpoint must be
    added; this looks up the previous header and raises MissingHeader if it cannot be
    found.  If the previous header is the tip of a chain the new header extends that chain
    and replaces its tip, otherwise a new chain is created with the header as its tip.

    Chains can only fork after the checkpoint header.  Headers before the checkpoint may
    be missing (i.e., the chain may have holes) and attempts to retrieve them raise a
    MissingHeader exception.  After the checkpoint all headers in a chain exist by
    construction.

    Headers can be looked up by height on a given chain.  They can be looked up by hash in
    which case the header and its chain are returned as a pair.

    Deserialized "Header" objects that are returnd always have their hash and height set
    in addition to the standard header attributes such as nonce and timestamp.
    '''

    max_cache_size = 1000

    def __init__(self, coin, storage, checkpoint):
        self.coin = coin
        self.checkpoint = checkpoint
        self._storage = storage
        self._chains = []
        self._short_hashes = bytearray()
        self._heights = array.array('I')
        self._chain_indices = array.array('H')
        self._cache = {}

        # Create the base chain out to the checkpoint
        self._add_chain(Chain.from_checkpoint(coin, checkpoint))
        # Read in chains from storage
        self._read_headers()

    def _add_chain(self, chain):
        chain.index = len(self._chains)
        self._chains.append(chain)
        self._add_chain_tip(chain)

    def _add_chain_tip(self, chain):
        header = chain.tip
        self._short_hashes.extend(header.hash[:4])
        self._heights.append(header.height)
        self._chain_indices.append(chain.index)
        # Add to cache; prevent it getting too big
        cache = self._cache
        cache[header.hash] = header, chain
        if len(cache) > self.max_cache_size:
            keys = list(cache.keys())
            for n in range(len(cache) // 2):
                del cache[keys[n]]
            cache.update({chain.tip.hash: (chain.tip, chain) for chain in self._chains})

    def _header_index_slow(self, header_hash):
        key = header_hash[:4]
        start = 0
        while True:
            index = self._short_hashes.find(key, start)
            if index % 4 == 0:
                our_index = index // 4
                raw_header = self._storage[our_index + self.checkpoint.height]
                if self.coin.header_hash(raw_header) == header_hash:
                    return raw_header, our_index
            if index == -1:
                raise MissingHeader(f'no header with hash {hash_to_hex_str(header_hash)}')
            start += 1

    def _read_headers(self):
        '''Read in all the headers from storage.'''
        read_header = self._read_header
        for header_index in range(self.checkpoint.height + 1, len(self)):
            read_header(header_index)

    def _read_header(self, header_index):
        '''Read a single header from storage.  The header must connect, either to extend an
        existing chain or create a new one.  Return the chain the header lies on.
        '''
        new_tip = self.coin.deserialized_header(self._storage[header_index], -1)
        prev_header, chain = self.lookup(new_tip.prev_hash)
        new_tip.height = prev_header.height + 1
        if chain.tip.hash == prev_header.hash:
            chain.append(new_tip, header_index)
            self._add_chain_tip(chain)
        else:
            prev_work = self.chainwork_to_height(chain, prev_header.height)
            chain = Chain(chain, new_tip, header_index, prev_work)
            self._add_chain(chain)
        return chain

    #
    # External API
    #

    @classmethod
    def from_file(cls, coin, file_path, checkpoint):
        storage = _HeaderStorage(file_path)
        storage.open_or_create(checkpoint)
        return cls(coin, storage, checkpoint)

    def set_one(self, height, raw_header):
        '''Set the raw header for a height before the checkpoint.

        The caller is responsible for the validity of the raw header.'''
        if not 0 <= height <= self.checkpoint.height:
            raise ValueError(f'cannot set header at height {height:,d}')
        self._storage[height] = raw_header

    def chainwork_range(self, chain, start_height, end_height):
        '''Returns the chainwork for the half-open range [start_height, end_height).'''
        raw_header = self._storage.__getitem__
        get_header_index = chain.header_index
        header_work = self.coin.header_work
        return sum(header_work(raw_header(get_header_index(h)))
                   for h in range(start_height, end_height))

    def chainwork_to_height(self, chain, height):
        '''Returns the chainwork to and including height on a chain.'''
        return chain.work - self.chainwork_range(chain, height + 1, chain.tip.height + 1)

    def raw_header_at_height(self, chain, height):
        return self._storage[chain.header_index(height)]

    def header_at_height(self, chain, height):
        raw_header = self.raw_header_at_height(chain, height)
        return self.coin.deserialized_header(raw_header, height)

    def lookup(self, header_hash):
        result = self._cache.get(header_hash)
        if result:
            return result
        raw_header, our_index = self._header_index_slow(header_hash)
        header = self.coin.deserialized_header(raw_header, self._heights[our_index])
        return header, self._chains[self._chain_indices[our_index]]

    def connect(self, raw_header):
        '''Given a raw header, try to connect it to existing headers.

        Returns a (header, chain) pair if the header could be connected or already exists.

        Check the header's "bits" and that the header's hash is good for the target that
        implies.

        Raises MissingHeader if the previous header cannot be found, IncorrectBits if the
        header's bits don't meet the chain's rules, and InsufficientPow if the header's
        hash doesn't meet the target.
        '''
        header = self.coin.deserialized_header(raw_header, -1)
        prev_header, chain = self.lookup(header.prev_hash)
        header.height = prev_header.height + 1
        # If the chain tip is the prior header then this header is new.  Otherwise we must
        # check.
        if chain.tip.hash != prev_header.hash:
            try:
                return self.lookup(header.hash)
            except MissingHeader:
                pass

        required_bits = self.required_bits(chain, header.height, header.timestamp)
        if header.bits != required_bits:
            raise IncorrectBits(header, required_bits)
        if header.hash_value() > header.target():
            raise InsufficientPoW(header)
        # OK, the header is good, store it and get its chain
        header_index = self._storage.append(raw_header)
        chain = self._read_header(header_index)
        return header, chain

    def __len__(self):
        '''The number of headers stored.'''
        return len(self._storage)

    def chains(self):
        return self._chains

    def chain_count(self):
        return len(self._chains)

    def longest_chain(self):
        longest = self._chains[0]
        for chain in self._chains:
            if chain.work > longest.work:
                longest = chain
        return longest

    def median_time_past(self, chain, height):
        '''Returns the median time past on chain at height.'''
        raw_header = self.raw_header_at_height
        timestamp = self.coin.header_timestamp
        timestamps = [timestamp(raw_header(chain, h))
                      for h in range(height, max(-1, height - 11), -1)]
        return sorted(timestamps)[len(timestamps) // 2]

    def required_bits(self, chain, height, timestamp=None):
        '''Returns the required bits for a new header at the given height with the
        given timestamp.  Testnet uses the timestamp; mainnet does not.'''
        return self.coin.required_bits(self, chain, height, timestamp)



# Copyright (c) 2018, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''Internal utilities'''

import mmap


def map_file(file_name, new_size=None):
    '''Map an existing file into memory.  If new_size is specified the
    file is truncated or extended to that size.

    Returns a Python mmap object.
    '''
    with open(file_name, 'rb+') as f:
        if new_size is not None:
            f.truncate(new_size)
        return mmap.mmap(f.fileno(), 0)


# Method decorator.  To be used for calculations that will always deliver the same result.
# The method cannot take any arguments and should be accessed as an attribute.
class cachedproperty(object):

    def __init__(self, f):
        self.f = f

    def __get__(self, obj, type_):
        obj = obj or type_
        value = self.f(obj)
        setattr(obj, self.f.__name__, value)
        return value



# Copyright (c) 2019, Neil Booth
# Copyright (C) 2018 The Electrum developers
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

'''AES wrapper.'''

__all__ = (
    'aes_encrypt_with_iv', 'aes_decrypt_with_iv', 'BadPaddingError',
)


from pyaes import AESModeOfOperationCBC, Decrypter, Encrypter, PADDING_NONE


class BadPaddingError(Exception):
    pass


def _append_PKCS7_padding(data):
    padlen = 16 - (len(data) % 16)
    return data + bytes([padlen]) * padlen


def _strip_PKCS7_padding(data):
    if not data or len(data) % 16:
        raise BadPaddingError('wrong length')
    padlen = data[-1]
    if not 0 < padlen <= 16:
        raise BadPaddingError('invalid final padding byte')
    if data[-padlen:] != bytes([padlen]) * padlen:
        raise BadPaddingError('inconsistent padding bytes')
    return data[:-padlen]


def aes_encrypt_with_iv(key, iv, data):
    data = _append_PKCS7_padding(data)
    aes_cbc = AESModeOfOperationCBC(key, iv=iv)
    aes = Encrypter(aes_cbc, padding=PADDING_NONE)
    return aes.feed(data) + aes.feed()  # empty aes.feed() flushes buffer


def aes_decrypt_with_iv(key, iv, data):
    aes_cbc = AESModeOfOperationCBC(key, iv=iv)
    aes = Decrypter(aes_cbc, PADDING_NONE)
    data = aes.feed(data) + aes.feed()  # empty aes.feed() flushes buffer
    return _strip_PKCS7_padding(data)



# Copyright (c) 2019, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''Miscellaneous functions.'''

__all__ = (
    'be_bytes_to_int', 'le_bytes_to_int',
    'int_to_be_bytes', 'int_to_le_bytes',
)

from functools import partial


# Converts big-endian bytes to an integer
be_bytes_to_int = partial(int.from_bytes, byteorder='big')
le_bytes_to_int = partial(int.from_bytes, byteorder='little')


def int_to_be_bytes(value, size=None):
    '''Converts an integer to a big-endian sequence of bytes'''
    if size is None:
        size = (value.bit_length() + 7) // 8
    return value.to_bytes(size, 'big')


def int_to_le_bytes(value, size=None):
    '''Converts an integer to a big-endian sequence of bytes'''
    if size is None:
        size = (value.bit_length() + 7) // 8
    return value.to_bytes(size, 'little')



# Copyright (c) 2019, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''Bitcoin script'''


__all__ = (
    'Ops', 'push_item', 'push_int', 'push_and_drop_item', 'push_and_drop_items', 'item_to_int',
    'Script', 'ScriptError', 'TruncatedScriptError',
)


from enum import IntEnum

from .misc import int_to_le_bytes, le_bytes_to_int
from .packing import (
    pack_byte, pack_le_uint16, pack_le_uint32,
    unpack_le_uint16, unpack_le_uint32,
)


class ScriptError(Exception):
    pass


class TruncatedScriptError(ScriptError):
    pass


SIGHASH_ALL = 0x01
SIGHASH_NONE = 0x02
SIGHASH_SINGLE = 0x03
SIGHASH_BASE_MASK = 0x1f
SIGHASH_FORKID = 0x40
SIGHASH_ANYONE_CAN_PAY = 0x080


class Ops(IntEnum):
    OP_0 = 0x00
    OP_FALSE = OP_0
    OP_PUSHDATA1 = 0x4c
    OP_PUSHDATA2 = 0x4d
    OP_PUSHDATA4 = 0x4e
    OP_1NEGATE = 0x4f
    OP_RESERVED = 0x50
    OP_1 = 0x51
    OP_TRUE = OP_1
    OP_2 = 0x52
    OP_3 = 0x53
    OP_4 = 0x54
    OP_5 = 0x55
    OP_6 = 0x56
    OP_7 = 0x57
    OP_8 = 0x58
    OP_9 = 0x59
    OP_10 = 0x5a
    OP_11 = 0x5b
    OP_12 = 0x5c
    OP_13 = 0x5d
    OP_14 = 0x5e
    OP_15 = 0x5f
    OP_16 = 0x60

    # control
    OP_NOP = 0x61
    OP_VER = 0x62
    OP_IF = 0x63
    OP_NOTIF = 0x64
    OP_VERIF = 0x65
    OP_VERNOTIF = 0x66
    OP_ELSE = 0x67
    OP_ENDIF = 0x68
    OP_VERIFY = 0x69
    OP_RETURN = 0x6a

    # stack ops
    OP_TOALTSTACK = 0x6b
    OP_FROMALTSTACK = 0x6c
    OP_2DROP = 0x6d
    OP_2DUP = 0x6e
    OP_3DUP = 0x6f
    OP_2OVER = 0x70
    OP_2ROT = 0x71
    OP_2SWAP = 0x72
    OP_IFDUP = 0x73
    OP_DEPTH = 0x74
    OP_DROP = 0x75
    OP_DUP = 0x76
    OP_NIP = 0x77
    OP_OVER = 0x78
    OP_PICK = 0x79
    OP_ROLL = 0x7a
    OP_ROT = 0x7b
    OP_SWAP = 0x7c
    OP_TUCK = 0x7d

    # splice ops
    OP_CAT = 0x7e
    OP_SPLIT = 0x7f    # after monolith upgrade (May 2018)
    OP_NUM2BIN = 0x80  # after monolith upgrade (May 2018)
    OP_BIN2NUM = 0x81  # after monolith upgrade (May 2018)
    OP_SIZE = 0x82

    # bit logic
    OP_INVERT = 0x83
    OP_AND = 0x84
    OP_OR = 0x85
    OP_XOR = 0x86
    OP_EQUAL = 0x87
    OP_EQUALVERIFY = 0x88
    OP_RESERVED1 = 0x89
    OP_RESERVED2 = 0x8a

    # numeric
    OP_1ADD = 0x8b
    OP_1SUB = 0x8c
    OP_2MUL = 0x8d
    OP_2DIV = 0x8e
    OP_NEGATE = 0x8f
    OP_ABS = 0x90
    OP_NOT = 0x91
    OP_0NOTEQUAL = 0x92

    OP_ADD = 0x93
    OP_SUB = 0x94
    OP_MUL = 0x95
    OP_DIV = 0x96
    OP_MOD = 0x97
    OP_LSHIFT = 0x98
    OP_RSHIFT = 0x99

    OP_BOOLAND = 0x9a
    OP_BOOLOR = 0x9b
    OP_NUMEQUAL = 0x9c
    OP_NUMEQUALVERIFY = 0x9d
    OP_NUMNOTEQUAL = 0x9e
    OP_LESSTHAN = 0x9f
    OP_GREATERTHAN = 0xa0
    OP_LESSTHANOREQUAL = 0xa1
    OP_GREATERTHANOREQUAL = 0xa2
    OP_MIN = 0xa3
    OP_MAX = 0xa4

    OP_WITHIN = 0xa5

    # crypto
    OP_RIPEMD160 = 0xa6
    OP_SHA1 = 0xa7
    OP_SHA256 = 0xa8
    OP_HASH160 = 0xa9
    OP_HASH256 = 0xaa
    OP_CODESEPARATOR = 0xab
    OP_CHECKSIG = 0xac
    OP_CHECKSIGVERIFY = 0xad
    OP_CHECKMULTISIG = 0xae
    OP_CHECKMULTISIGVERIFY = 0xaf

    # expansion
    OP_NOP1 = 0xb0
    OP_CHECKLOCKTIMEVERIFY = 0xb1
    OP_NOP2 = OP_CHECKLOCKTIMEVERIFY
    OP_CHECKSEQUENCEVERIFY = 0xb2
    OP_NOP3 = OP_CHECKSEQUENCEVERIFY
    OP_NOP4 = 0xb3
    OP_NOP5 = 0xb4
    OP_NOP6 = 0xb5
    OP_NOP7 = 0xb6
    OP_NOP8 = 0xb7
    OP_NOP9 = 0xb8
    OP_NOP10 = 0xb9


globals().update((f'b_{name}', pack_byte(value)) for name, value in Ops.__members__.items())
globals().update(Ops.__members__)
__all__ += tuple(f'b_{name}' for name in Ops.__members__.keys())
__all__ += tuple(Ops.__members__.keys())


b_OP_DUP_OP_HASH160 = bytes([OP_DUP, OP_HASH160])
b_OP_EQUALVERIFY_OP_CHECKSIG = bytes([OP_EQUALVERIFY, OP_CHECKSIG])


def push_item(item):
    '''Returns script bytes to push item on the stack.'''
    dlen = len(item)
    if dlen <= 1:
        # Values 1...16 and 0x81 can be pushed specially as a single opcode.
        if dlen == 0:
            return b_OP_0
        value = item[0]
        if 0 < value <= 16:
            return pack_byte(OP_1 + value - 1)
        if value == 0x81:
            return b_OP_1NEGATE

    if dlen < OP_PUSHDATA1:
        return pack_byte(dlen) + item
    if dlen <= 0xff:
        return pack_byte(OP_PUSHDATA1) + pack_byte(dlen) + item
    if dlen <= 0xffff:
        return pack_byte(OP_PUSHDATA2) + pack_le_uint16(dlen) + item
    return pack_byte(OP_PUSHDATA4) + pack_le_uint32(dlen) + item


def push_and_drop_item(item):
    '''Push one item onto the stack and then pop it off.'''
    return push_item(item) + b_OP_DROP


def push_and_drop_items(items):
    '''Push several items onto the stack and then pop them all off.'''
    parts = [push_item(item) for item in items]
    if len(items) >= 2:
        parts.append(b_OP_2DROP * (len(parts) // 2))
    if len(items) & 1:
        parts.append(b_OP_DROP)
    return b''.join(parts)


def push_int(value):
    '''Returns script bytes to push a numerical value to the stack.  Stack values are stored as
    signed-magnitude little-endian numbers.
    '''
    if value == 0:
        return b_OP_0
    item = int_to_le_bytes(abs(value))
    if item[-1] & 0x80:
        item += pack_byte(0x80 if value < 0 else 0x00)
    elif value < 0:
        item = item[:-1] + pack_byte(item[-1] | 0x80)
    return push_item(item)


def item_to_int(item):
    '''Returns the value of a stack item interpreted as an integer.'''
    if not item:
        return 0
    if item[-1] & 0x80:
        return -le_bytes_to_int(item[:-1] + pack_byte(item[-1] & 0x7f))
    return le_bytes_to_int(item)


class Script(bytes):
    '''Wraps the raw bytes of a bitcoin script.'''

    def ops(self):
        '''A generator.  Yields script operations (does not check their validity) as integers, and
        data pushed on the stack as bytes.  Returns when the end of the script is reached.

        Raises TruncatedScriptError if the script was truncated.
        '''
        n = 0
        limit = len(self)

        while n < limit:
            op = self[n]
            n += 1
            if op > OP_16:
                yield op
            elif op <= OP_PUSHDATA4:
                try:
                    if op < OP_PUSHDATA1:
                        dlen = op
                    elif op == OP_PUSHDATA1:
                        dlen = self[n]
                        n += 1
                    elif op == OP_PUSHDATA2:
                        dlen, = unpack_le_uint16(self[n: n + 2])
                        n += 2
                    else:
                        dlen, = unpack_le_uint32(self[n: n + 4])
                        n += 4
                    data = self[n: n + dlen]
                    n += dlen
                    assert len(data) == dlen
                    yield data
                except Exception:
                    raise TruncatedScriptError from None
            elif op >= OP_1:
                yield pack_byte(op - OP_1 + 1)
            elif op == OP_1NEGATE:
                yield b'\x81'
            else:
                assert op == OP_RESERVED
                yield op

    @classmethod
    def op_to_asm_word(cls, op):
        '''Convert a single opcode, or data push, as returned by ops(), to a human-readable
        word.'''
        if isinstance(op, bytes):
            if len(op) <= 4:
                return str(item_to_int(op))
            # Should we implement bitcoind's sigHashDecode?
            return op.hex()
        try:
            return Ops(op).name
        except ValueError:
            return "OP_INVALIDOPCODE" if op == 0xff else "OP_UNKNOWN"

    def to_asm(self):
        '''Return a script converted to bitcoin's human-readable ASM format.'''
        op_to_asm_word = self.op_to_asm_word
        try:
            return ' '.join(op_to_asm_word(op) for op in self.ops())
        except TruncatedScriptError:
            return '[error]'

    def to_bytes(self):
        '''Return the script as a bytes() object.'''
        return self

    @classmethod
    def from_hex(cls, hex_str):
        '''Instantiate from a hexadecimal string.'''
        return cls(bytes.fromhex(hex_str))

    def to_hex(self):
        '''Return the script as a hexadecimal string.'''
        return self.hex()

    @classmethod
    def asm_word_to_bytes(cls, word):
        '''Convert an ASM word to bytes, either a 1-byte opcode or the data bytes.'''
        if word.startswith('OP_'):
            try:
                opcode = Ops[word]
            except KeyError:
                raise ScriptError(f'unrecognized op code {word}') from None
            return pack_byte(opcode)
        # Handle what looks like a decimal, provided it's in-range
        if word.isdigit() or word[0] == '-' and word[1:].isdigit():
            value = int(word)
            if abs(value) <= 2147483647:
                return push_int(value)
        try:
            return push_item(bytes.fromhex(word))
        except ValueError:
            raise ScriptError(f'invalid pushdata {word}') from None

    @classmethod
    def from_asm(cls, asm):
        '''Convert an ASM string to a script.'''
        asm_word_to_bytes = cls.asm_word_to_bytes
        return cls(b''.join(asm_word_to_bytes(word) for word in asm.split()))

    @classmethod
    def P2PK_script(cls, pubkey_bytes):
        if len(pubkey_bytes) not in (33, 65):
            raise ValueError('invalid pubkey_bytes')
        return cls(push_item(pubkey_bytes) + b_OP_CHECKSIG)

    @classmethod
    def P2PKH_script(cls, hash160_bytes):
        if len(hash160_bytes) != 20:
            raise ValueError('invalid hash160_bytes')
        return cls(b''.join((b_OP_DUP_OP_HASH160, push_item(hash160_bytes),
                             b_OP_EQUALVERIFY_OP_CHECKSIG)))

    @classmethod
    def P2SH_script(cls, hash160_bytes):
        if len(hash160_bytes) != 20:
            raise ValueError('invalid hash160_bytes')
        return cls(b''.join((b_OP_HASH160, push_item(hash160_bytes), b_OP_EQUAL)))



# Copyright (c) 2018, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# and warranty status of this software.

__all__ = (
    'pack_le_int32', 'pack_le_int64',
    'pack_le_uint16', 'pack_le_uint32', 'pack_le_uint64',
    'pack_be_uint16', 'pack_be_uint32',
    'pack_byte', 'pack_port', 'pack_varint', 'pack_varbytes', 'pack_list',
    'unpack_le_int32', 'unpack_le_int32_from',
    'unpack_le_int64', 'unpack_le_int64_from',
    'unpack_le_uint16', 'unpack_le_uint16_from',
    'unpack_le_uint32', 'unpack_le_uint32_from',
    'unpack_le_uint64', 'unpack_le_uint64_from',
    'unpack_be_uint16', 'unpack_be_uint16_from',
    'unpack_be_uint32', 'unpack_be_uint32_from',
    'unpack_byte', 'unpack_port', 'unpack_header',
    'read_le_int32', 'read_le_int64',
    'read_le_uint16', 'read_le_uint32', 'read_le_uint64',
    'read_be_uint16', 'read_be_uint32',
    'read_varint', 'read_varbytes', 'read_list',
)


from struct import Struct, error as struct_error


struct_le_i = Struct('<i')
struct_le_q = Struct('<q')
struct_le_H = Struct('<H')
struct_le_I = Struct('<I')
struct_le_Q = Struct('<Q')
struct_be_H = Struct('>H')
struct_be_I = Struct('>I')
structB = Struct('B')
struct_header = Struct('<i 32s 32s 3I')

pack_le_int32 = struct_le_i.pack
pack_le_int64 = struct_le_q.pack
pack_le_uint16 = struct_le_H.pack
pack_le_uint32 = struct_le_I.pack
pack_le_uint64 = struct_le_Q.pack
pack_be_uint16 = struct_be_H.pack
pack_be_uint32 = struct_be_I.pack
pack_byte = structB.pack

unpack_le_int32 = struct_le_i.unpack
unpack_le_int32_from = struct_le_i.unpack_from
unpack_le_int64 = struct_le_q.unpack
unpack_le_int64_from = struct_le_q.unpack_from
unpack_le_uint16 = struct_le_H.unpack
unpack_le_uint16_from = struct_le_H.unpack_from
unpack_le_uint32 = struct_le_I.unpack
unpack_le_uint32_from = struct_le_I.unpack_from
unpack_le_uint64 = struct_le_Q.unpack
unpack_le_uint64_from = struct_le_Q.unpack_from
unpack_be_uint16 = struct_be_H.unpack
unpack_be_uint16_from = struct_be_H.unpack_from
unpack_be_uint32 = struct_be_I.unpack
unpack_be_uint32_from = struct_be_I.unpack_from
unpack_byte = structB.unpack
unpack_header = struct_header.unpack

pack_port = pack_be_uint16
unpack_port = unpack_be_uint16
hex_to_bytes = bytes.fromhex


def pack_varint(n):
    '''Convert an unsigned integer into a binary varint (CompactSize).

    Return a bytes object.'''
    if n < 253:
        return pack_byte(n)
    if n < 65536:
        return b'\xfd' + pack_le_uint16(n)
    if n < 4294967296:
        return b'\xfe' + pack_le_uint32(n)
    return b'\xff' + pack_le_uint64(n)


def pack_varbytes(data):
    '''Serialize binary data by prepending a size varint.'''
    return pack_varint(len(data)) + data


def pack_list(items, pack_one):
    '''Pack a list of items.

    Each item is packed with pack_one, the stream begins with the item count.'''
    parts = [pack_varint(len(items))]
    parts.extend(pack_one(item) for item in items)
    return b''.join(parts)


# Stream operations


def read_le_int32(read):
    result, = unpack_le_int32(read(4))
    return result


def read_le_int64(read):
    result, = unpack_le_int64(read(8))
    return result


def read_le_uint16(read):
    result, = unpack_le_uint16(read(2))
    return result


def read_le_uint32(read):
    result, = unpack_le_uint32(read(4))
    return result


def read_le_uint64(read):
    result, = unpack_le_uint64(read(8))
    return result


def read_be_uint16(read):
    result, = unpack_be_uint16(read(2))
    return result


def read_be_uint32(read):
    result, = unpack_be_uint32(read(4))
    return result


def read_varint(read):
    # read_byte is supported by mmap objects but not BytesIO
    n, = unpack_byte(read(1))
    if n < 253:
        return n
    if n == 253:
        return read_le_uint16(read)
    if n == 254:
        return read_le_uint32(read)
    return read_le_uint64(read)


def read_varbytes(read):
    n = read_varint(read)
    result = read(n)
    if len(result) != n:
        raise struct_error(f'varbytes requires a buffer of {n:,d} bytes')
    return result


def read_list(read, read_one):
    '''Return a list of items.

    Each item is read with read_one, the stream begins with a count of the items.'''
    return [read_one(read) for _ in range(read_varint(read))]



# Copyright (c) 2018, 2019 Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

__all__ = (
    'bits_to_target', 'target_to_bits', 'bits_to_work',
)

from functools import lru_cache


@lru_cache()
def bits_to_target(bits):
    if not 0 <= bits <= 0x2100ffff:
        raise ValueError(f'bits value 0x{bits:x} out of range')

    word = bits & 0x00ffffff
    # target_to_bits never generates these values
    if (not 0x8000 <= word <= 0x7fffff) and bits:
        raise ValueError(f'bits value 0x{bits:x} is invalid')

    size = bits >> 24
    shift = 8 * (size - 3)
    if shift <= 0:
        result = word >> -shift
        # target_to_bits never generates these values
        if (result << -shift) != word:
            raise ValueError(f'bits value 0x{bits:x} is invalid')
        return result
    else:
        return word << shift


def target_to_bits(target):
    bits = target.bit_length()
    if target < 0 or bits > 256:
        raise ValueError(f'target 0x{target:x} out of range')

    size = (bits + 7) // 8
    shift = 8 * (size - 3)
    if shift <= 0:
        word = target << -shift
    else:
        word = target >> shift

    # Avoid setting the sign bit
    if word & 0x00800000:
        word >>= 8
        size += 1

    return word | (size << 24)


@lru_cache()
def bits_to_work(bits):
    return (1 << 256) // (bits_to_target(bits) + 1)


def _required_bits_fortnightly(headers, chain, height):
    '''Bitcoin's original DAA.'''
    if height == 0:
        return headers.coin.genesis_bits
    prev = headers.header_at_height(chain, height - 1)
    if height % 2016:
        return prev.bits
    prior = headers.header_at_height(chain, height - 2016)

    # Off-by-one with prev.timestamp.  Constrain the actual time.
    period = prev.timestamp - prior.timestamp
    target_period = 2016 * 600
    adj_period = min(max(period, target_period // 4), target_period * 4)

    prior_target = bits_to_target(prev.bits)
    new_target = (prior_target * adj_period) // target_period
    return target_to_bits(min(new_target, headers.coin.max_target))


def _required_bits_DAA(headers, chain, height):
    '''BCH's shoddy difficulty adjustment algorithm.  He was warned, he shrugged.'''
    def median_prior_header(ref_height):
        '''Select the median of the 3 prior headers, for a curious definition of median.'''
        def maybe_swap(m, n):
            if prev3[m].timestamp > prev3[n].timestamp:
                prev3[m], prev3[n] = prev3[n], prev3[m]

        prev3 = [header_at_height(chain, h) for h in range(ref_height - 3, ref_height)]
        maybe_swap(0, 2)
        maybe_swap(0, 1)
        maybe_swap(1, 2)
        return prev3[1]

    header_at_height = headers.header_at_height
    start = median_prior_header(height - 144)
    end = median_prior_header(height)

    period_work = headers.chainwork_range(chain, start.height + 1, end.height + 1)
    period_time = min(max(end.timestamp - start.timestamp, 43200), 172800)

    Wn = (period_work * 600) // period_time
    new_target = (1 << 256) // Wn - 1
    return target_to_bits(min(new_target, headers.coin.max_target))


def _required_bits_EDA(headers, chain, height):
    '''The less said the better.'''
    bits = _required_bits_fortnightly(headers, chain, height)
    if height % 2016 == 0:
        return bits

    mtp_diff = (headers.median_time_past(chain, height - 1) -
                headers.median_time_past(chain, height - 7))
    if mtp_diff < 12 * 3600:
        return bits

    # Increase target by 25% (reducing difficulty by 20%).
    new_target = bits_to_target(bits)
    new_target += new_target >> 2
    return target_to_bits(min(new_target, headers.coin.max_target))


def required_bits_mainnet(headers, chain, height, _timestamp=None):
    # Unlike testnet, required_bits is not a function of the timestamp
    if height < 478558:
        return _required_bits_fortnightly(headers, chain, height)
    elif height <= 504031:
        return _required_bits_EDA(headers, chain, height)
    else:
        return _required_bits_DAA(headers, chain, height)


def _required_bits_testnet(headers, chain, height, timestamp, daa_height):
    def prior_non_special_bits():
        genesis_bits = headers.coin.genesis_bits
        raw_header = headers.raw_header_at_height
        header_bits = headers.coin.header_bits
        for test_height in range(height - 1, -1, -1):
            bits = header_bits(raw_header(chain, test_height))
            if test_height % 2016 == 0 or bits != genesis_bits:
                return bits
        # impossible to fall through here

    prior_raw_header = headers.raw_header_at_height(chain, height - 1)
    prior_timestamp = headers.coin.header_timestamp(prior_raw_header)
    is_slow = (timestamp - prior_timestamp) > 20 * 60

    if height <= daa_height:
        # Note: testnet did not use the EDA
        if height % 2016 == 0:
            return _required_bits_fortnightly(headers, chain, height)
        if is_slow:
            return headers.coin.genesis_bits
        return prior_non_special_bits()
    else:
        if is_slow:
            return headers.coin.genesis_bits
        return _required_bits_DAA(headers, chain, height)


def required_bits_testnet(headers, chain, height, timestamp):
    return _required_bits_testnet(headers, chain, height, timestamp, 1188697)


def required_bits_scaling_testnet(headers, chain, height, timestamp):
    return _required_bits_testnet(headers, chain, height, timestamp, 2200)



import os.path
import re
import setuptools


def find_version(filename):
    with open(filename) as f:
        text = f.read()
    match = re.search(r"^_version_str = '(.*)'$", text, re.MULTILINE)
    if not match:
        raise RuntimeError('cannot find version')
    return match.group(1)


tld = os.path.abspath(os.path.dirname(__file__))
version = find_version(os.path.join(tld, 'bitcoinx', '__init__.py'))


setuptools.setup(
    name='bitcoinX',
    version=version,
    python_requires='>=3.6',
    install_requires=['attrs', 'pyaes', 'electrumsv-secp256k1'],
    packages=['bitcoinx'],
    description='Library of Bitcoin functions',
    author='Neil Booth',
    author_email='kyuupichan@gmail.com',
    license='MIT Licence',
    url='https://github.com/kyuupichan/bitcoinX',
    download_url=('https://github.com/kyuupichan/bitcoinX/archive/'
                  f'{version}.tar.gz'),
    long_description=(
        'Library of Bitcoin functions covering network protocol, consensus, '
        'transactions, scripting and signing.'
    ),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        "Programming Language :: Python :: 3.6",
        'Topic :: Internet',
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



# Copyright (c) 2019, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''BIP32 implementation.'''

__all__ = (
    'BIP32PublicKey', 'BIP32PrivateKey', 'BIP32Derivation',
    'bip32_key_from_string', 'bip32_decompose_chain_string', 'bip32_is_valid_chain_string',
)

from os import urandom
import re

import attr

from .base58 import base58_decode_check, base58_encode_check
from .coin import Bitcoin, Coin
from .hashes import hmac_sha512_halves, hash160
from .keys import PrivateKey, PublicKey
from .packing import pack_be_uint32, unpack_be_uint32, pack_byte
from .util import cachedproperty

HARDENED = 1 << 31
PART_REGEX = re.compile("([0-9]+)'?$")


@attr.s(slots=True, frozen=True, repr=False)
class BIP32Derivation(object):
    '''Metadata about a BIP32 derivation.'''
    chain_code = attr.ib()
    n = attr.ib()
    depth = attr.ib()
    parent_fingerprint = attr.ib()

    def extended_key(self, coin, raw_serkey):
        '''Return the 78-byte extended key bytes.'''
        if len(raw_serkey) == 32:
            raw_serkey = b'\0' + raw_serkey
            ver_bytes = coin.xprv_verbytes
        else:
            ver_bytes = coin.xpub_verbytes

        assert len(raw_serkey) == 33

        return b''.join((
            ver_bytes,
            pack_byte(self.depth),
            self.parent_fingerprint,
            pack_be_uint32(self.n),
            self.chain_code,
            raw_serkey,
        ))

    def child(self, chain_code, n, parent_fingerprint):
        return BIP32Derivation(chain_code, n, self.depth + 1, parent_fingerprint)

    def __repr__(self):
        return (f'BIP32Derivation(chain_code=bytes.fromhex("{self.chain_code.hex()}"), '
                f'n={self.n}, depth={self.depth}, '
                f'parent_fingerprint=bytes.fromhex("{self.parent_fingerprint.hex()}"))')


class BIP32PrivateKey(PrivateKey):
    '''A BIP32 private key.

    Intended to be constructed in one of the following ways:
       - from a string via bip32_key_from_string()
       - the classmethod from_seed
       - from an existing instance with the child() function
    '''
    def __init__(self, secret, derivation, coin):
        super().__init__(secret, True, coin)
        self._derivation = derivation

    def _extended_key(self, coin):
        '''Return a raw extended private key.'''
        coin = coin or self._coin
        return self._derivation.extended_key(coin, self._secret)

    @classmethod
    def _from_parts(cls, privkey, chain_code, coin):
        derivation = BIP32Derivation(chain_code, 0, 0, bytes(4))
        return cls(privkey, derivation, coin)

    @classmethod
    def from_seed(cls, seed, *, coin=Bitcoin):
        # This hard-coded message string seems to be coin-independent...
        privkey, chain_code = hmac_sha512_halves(b'Bitcoin seed', seed)
        return cls._from_parts(privkey, chain_code, coin)

    @classmethod
    def from_random(cls, *, source=urandom):
        '''Return a random, valid PrivateKey.'''
        while True:
            try:
                data = source(64)
                return cls._from_parts(data[:32], data[32:], Bitcoin)
            except ValueError:
                pass

    @cachedproperty
    def public_key(self):
        '''Return the corresponding BIP32PublicKey object.'''
        return BIP32PublicKey(self._secp256k1_public_key(), self._derivation, self._coin)

    def child(self, n):
        '''Return the derived child extended privkey at index N.'''
        if not 0 <= n < (1 << 32):
            raise ValueError(f'invalid BIP32 private key child number: {n}')

        if n >= HARDENED:
            serkey = b'\0' + self._secret
        else:
            serkey = self.public_key.to_bytes()

        msg = serkey + pack_be_uint32(n)
        L, R = hmac_sha512_halves(self._derivation.chain_code, msg)
        child_derivation = self._derivation.child(R, n, self.fingerprint())

        return BIP32PrivateKey(self.add(L).to_bytes(), child_derivation, self._coin)

    def child_safe(self, n):
        '''Return a child but increment n if the child derivation is invalid.'''
        while True:
            try:
                return self.child(n)
            except ValueError:
                if not 0 <= n < (1 << 32):
                    raise
                n += 1
                if n & (HARDENED - 1) == 0:
                    raise ValueError('out of BIP32 derivations')

    def derivation(self):
        '''Return a BIP32 derivation object.'''
        return self._derivation

    def identifier(self):
        '''Return the key's identifier as 20 bytes.'''
        return self.public_key.identifier()

    def fingerprint(self):
        '''Return the key's fingerprint as 4 bytes.'''
        return self.public_key.fingerprint()

    def to_extended_key_string(self, *, coin=None):
        '''Return an extended key as a base58 string.'''
        return base58_encode_check(self._extended_key(coin))

    # Remove soon
    extended_key_string = to_extended_key_string

    def __repr__(self):
        return f'BIP32PrivateKey("{str(self)}")'


class BIP32PublicKey(PublicKey):
    '''A BIP32 public key.

    Intended to be constructed in one of the following ways:
       - from a string via bip32_key_from_string()
       - from an existing instance with the child() function
       - from a BIP32PrivateKey via its public_key attribute
    '''

    def __init__(self, public_key, derivation, coin):
        if isinstance(public_key, PublicKey):
            public_key = public_key._public_key
        super().__init__(public_key, True, coin)
        self._derivation = derivation

    def _extended_key(self, coin):
        '''Return a raw extended private key.'''
        coin = coin or self._coin
        return self._derivation.extended_key(coin, self.to_bytes())

    def child(self, n):
        '''Return the derived child extended pubkey at index N.'''
        if not 0 <= n < HARDENED:
            raise ValueError(f'invalid BIP32 public key child number: {n}')

        msg = self.to_bytes() + pack_be_uint32(n)
        L, R = hmac_sha512_halves(self._derivation.chain_code, msg)
        child_derivation = self._derivation.child(R, n, self.fingerprint())

        return BIP32PublicKey(self.add(L), child_derivation, self._coin)

    def child_safe(self, n):
        '''Return a child but increment n if the child derivation is invalid.'''
        while True:
            try:
                return self.child(n)
            except ValueError:
                if not 0 <= n < HARDENED:
                    raise
                n += 1

    def derivation(self):
        '''Return a BIP32 derivation object.'''
        return self._derivation

    def identifier(self):
        '''Return the key's identifier as 20 bytes.'''
        return hash160(self.to_bytes())

    def fingerprint(self):
        '''Return the key's fingerprint as 4 bytes.'''
        return self.identifier()[:4]

    def to_extended_key_string(self, *, coin=None):
        '''Return an extended key as a base58 string.'''
        return base58_encode_check(self._extended_key(coin))

    # Remove soon
    extended_key_string = to_extended_key_string

    def __str__(self):
        return self.to_extended_key_string()

    def __repr__(self):
        return f'BIP32PublicKey("{self.to_extended_key_string()}")'


def _from_extended_key(ekey):
    '''Return a PubKey or PrivKey from an extended key raw bytes.'''
    if len(ekey) != 78:
        raise ValueError('extended key must have length 78')

    coin, is_public_key = Coin.lookup_xver_bytes(ekey[:4])
    n, = unpack_be_uint32(ekey[9:13])
    derivation = BIP32Derivation(ekey[13:45], n, ekey[4], ekey[5:9])
    if is_public_key:
        key = BIP32PublicKey(PublicKey.from_bytes(ekey[45:]), derivation, coin)
    else:
        if ekey[45] != 0:
            raise ValueError(f'invalid extended private key prefix byte {ekey[45]}')
        key = BIP32PrivateKey(ekey[46:], derivation, coin)

    return key


def bip32_key_from_string(ekey_str):
    '''Given an extended key string, such as

    xpub6BsnM1W2Y7qLMiuhi7f7dbAwQZ5Cz5gYJCRzTNainXzQXYjFwtuQXHd
    3qfi3t3KJtHxshXezfjft93w4UE7BGMtKwhqEHae3ZA7d823DVrL

    return a BIP32PubKey or BIP32PrivKey.
    '''
    return _from_extended_key(base58_decode_check(ekey_str))


def bip32_decompose_chain_string(chain_str):
    '''Given a chain string return a list of unsigned integers.

       For example:  m/1/2'/3'/0  -> [1, 0x80000002, 0x800000003, 0]
                     m            -> []

       The chain string must be 'm' or begin with 'm/'.
    '''
    if not isinstance(chain_str, str):
        raise TypeError(f'chain_str {chain_str} must be a string')

    parts = chain_str.split('/')
    if not parts or parts[0] != 'm':
        raise ValueError(f'invalid bip32 chain: {chain_str}')

    result = []
    for part in parts[1:]:
        match = PART_REGEX.match(part)
        if not match:
            raise ValueError(f'invalid bip32 chain: {chain_str}')
        value = int(match.groups()[0])
        if value >= 2147483648:
            raise ValueError(f'invalid bip32 chain: {chain_str}')
        if part[-1] == "'":
            value += HARDENED
        result.append(value)
    return result


def bip32_is_valid_chain_string(chain_str):
    '''Return True if chain_str is a valid BIP32 chain string.'''
    try:
        bip32_decompose_chain_string(chain_str)
        return True
    except ValueError:
        return False



# Copyright (c) 2019, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


__all__ = (
    'Tx', 'TxInput', 'TxOutput',
)

import attr
from io import BytesIO
from itertools import chain

from .hashes import hash_to_hex_str, double_sha256
from .packing import (
    pack_le_int32, pack_le_uint32, pack_varint, pack_varbytes, pack_le_int64, pack_list,
    read_le_int32, read_le_uint32, read_varint, read_varbytes, read_le_int64, read_list,
)
from .script import (
    SIGHASH_ALL, SIGHASH_NONE, SIGHASH_SINGLE, SIGHASH_ANYONE_CAN_PAY, SIGHASH_FORKID,
    SIGHASH_BASE_MASK
)

ZERO = bytes(32)
MINUS_1 = 4294967295


@attr.s(slots=True)
class Tx:
    '''A bitcoin transaction.'''
    version = attr.ib()
    inputs = attr.ib()
    outputs = attr.ib()
    locktime = attr.ib()

    def is_coinbase(self):
        '''Return True iff the tx is a coinbase transaction.'''
        return self.inputs[0].is_coinbase()

    @classmethod
    def read(cls, read):
        return cls(
            read_le_int32(read),
            read_list(read, TxInput.read),
            read_list(read, TxOutput.read),
            read_le_uint32(read),
        )

    @classmethod
    def from_bytes(cls, raw):
        return cls.read(BytesIO(raw).read)

    def to_bytes(self):
        return b''.join((
            pack_le_int32(self.version),
            pack_list(self.inputs, TxInput.to_bytes),
            pack_list(self.outputs, TxOutput.to_bytes),
            pack_le_uint32(self.locktime),
        ))

    @classmethod
    def from_hex(cls, hex_str):
        return cls.from_bytes(bytes.fromhex(hex_str))

    def to_hex(self):
        return self.to_bytes().hex()

    def _hash_prevouts(self):
        preimage = b''.join(txin.prevout_bytes() for txin in self.inputs)
        return double_sha256(preimage)

    def _hash_sequence(self):
        preimage = b''.join(pack_le_uint32(txin.sequence) for txin in self.inputs)
        return double_sha256(preimage)

    def _hash_outputs(self):
        preimage = b''.join(txout.to_bytes() for txout in self.outputs)
        return double_sha256(preimage)

    def signature_hash(self, input_index, value, script, *, sighash=SIGHASH_ALL | SIGHASH_FORKID):
        if not 0 <= input_index < len(self.inputs):
            raise IndexError(f'invalid input index: {input_index}')
        if value < 0:
            raise ValueError(f'value cannot be negative: {value}')

        txin = self.inputs[input_index]
        hash_prevouts = hash_sequence = hash_outputs = ZERO

        sighash_base = sighash & SIGHASH_BASE_MASK
        sighash_not_single_none = sighash_base not in (SIGHASH_SINGLE, SIGHASH_NONE)
        if not (sighash & SIGHASH_ANYONE_CAN_PAY):
            hash_prevouts = self._hash_prevouts()
            if sighash_not_single_none:
                hash_sequence = self._hash_sequence()
        if sighash_not_single_none:
            hash_outputs = self._hash_outputs()
        elif (sighash_base == SIGHASH_SINGLE and input_index < len(self.outputs)):
            hash_outputs = double_sha256(self.outputs[input_index].to_bytes())

        preimage = b''.join((
            pack_le_int32(self.version),
            hash_prevouts,
            hash_sequence,
            txin.to_bytes_for_signature(value, script),
            hash_outputs,
            pack_le_uint32(self.locktime),
            pack_le_uint32(sighash),
        ))

        return double_sha256(preimage)


@attr.s(slots=True, repr=False)
class TxInput:
    '''A bitcoin transaction input.'''
    prev_hash = attr.ib()
    prev_idx = attr.ib()
    script_sig = attr.ib()
    sequence = attr.ib()

    def is_coinbase(self):
        '''Return True iff the input is the single input of a coinbase transaction.'''
        return self.prev_idx == MINUS_1 and self.prev_hash == ZERO

    @classmethod
    def read(cls, read):
        return cls(
            read(32),               # prev_hash
            read_le_uint32(read),   # prev_idx
            read_varbytes(read),    # script_sig
            read_le_uint32(read),   # sequence
        )

    def prevout_bytes(self):
        return self.prev_hash + pack_le_uint32(self.prev_idx)

    def to_bytes_for_signature(self, value, script):
        return b''.join((
            self.prevout_bytes(),
            pack_varbytes(script),
            pack_le_int64(value),
            pack_le_uint32(self.sequence),
        ))

    def to_bytes(self):
        '''Pass value to get a serialization to be used in transaction signatures.'''
        return b''.join((
            self.prevout_bytes(),
            pack_varbytes(self.script_sig),
            pack_le_uint32(self.sequence),
        ))

    def __repr__(self):
        return (
            f'TxInput(prev_hash="{hash_to_hex_str(self.prev_hash)}", prev_idx={self.prev_idx}, '
            f'script_sig="{self.script_sig.hex()}", sequence={self.sequence})'
        )


@attr.s(slots=True, repr=False)
class TxOutput:
    '''A bitcoin transaction output.'''
    value = attr.ib()
    script_pk = attr.ib()

    @classmethod
    def read(cls, read):
        return cls(
            read_le_int64(read),   # value
            read_varbytes(read),   # script_pk
        )

    def to_bytes(self):
        return b''.join((
            pack_le_int64(self.value),
            pack_varbytes(self.script_pk),
        ))

    def __repr__(self):
        return (
            f'TxOutput(value={self.value}, script_pk="{self.script_pk.hex()}")'
        )
